"""This file provides the ability to combine vulnerabilities from multiple sources into a single model"""
# This file is part of hoppr-cop
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Copyright (c) 2022 Lockheed Martin Corporation
from typing import List, Optional, Dict

from hoppr_cyclonedx_models.cyclonedx_1_4 import Vulnerability


def __get_all_ids_from_vulnerability(from_vulnerability: Vulnerability):
    ids = [from_vulnerability.id]
    if from_vulnerability.references is not None:
        for reference in from_vulnerability.references:
            ids.append(reference.id)
    return list(set(ids))


def __combine_vunerabilities(
    existing: Vulnerability, new: Vulnerability
) -> Vulnerability:

    # pylint: disable=cell-var-from-loop

    if existing.cwes is None or len(existing.cwes) == 0:
        existing.cwes = new.cwes

    ## Combine advisories
    if existing.advisories is None and new.advisories is not None:
        existing.advisories = new.advisories
    elif new.advisories is not None:
        existing_urls = list(map(lambda x: x.url, existing.advisories))
        new_advisories = list(
            filter(lambda x: x.url not in existing_urls, new.advisories)
        )
        existing.advisories = existing.advisories + new_advisories

    # Combine Tools
    existing.tools = [] if existing.tools is None else existing.tools
    if new.tools is not None:
        existing_tool_names = list(map(lambda x: x.name, existing.tools))
        new_tools = list(filter(lambda x: x.name not in existing_tool_names, new.tools))
        existing.tools = new_tools + existing.tools

    # Combine Ratings
    existing.ratings = [] if existing.ratings is None else existing.ratings
    if new.ratings is not None:
        existing_reference_ids = list(map(lambda x: str(x.vector), existing.ratings))
        new_ratings = list(
            filter(lambda x: str(x.vector) not in existing_reference_ids, new.ratings)
        )
        existing.ratings = new_ratings + existing.ratings
        for rating in new.ratings:
            if rating.source is not None:
                existing_rating = list(
                    filter(lambda x: x.vector == rating.vector, existing.ratings)
                )[0]
                if (
                    existing_rating.source is None
                    or existing_rating.source.name is None
                ):
                    existing_rating.source = rating.source

    # Combine References
    existing.references = [] if existing.references is None else existing.references
    if new.references is not None:
        existing_reference_ids = list(map(lambda x: str(x.id), existing.references))
        new_references = list(
            filter(lambda x: str(x.id) not in existing_reference_ids, new.references)
        )
        existing.references = new_references + existing.references

    # Combine Recommendation
    existing.recommendation = (
        "" if existing.recommendation is None else existing.recommendation
    )
    if new.recommendation is not None and new.recommendation != "":
        existing.recommendation = (
            f"{existing.recommendation}.  ||  {new.recommendation}"
        )

    return existing


def __does_id_exist_in_both(
    existing_id_list: List[str], new_id_list: List[str]
) -> bool:
    length = len([value for value in existing_id_list if value in new_id_list]) > 0

    return length


def __get_matching_vulnerability_from_list(
    combined_list: List[Vulnerability], new_finding: Vulnerability
) -> Optional[Vulnerability]:
    new_finding_ids = __get_all_ids_from_vulnerability(new_finding)
    for vuln in combined_list:
        existing_finding_ids = __get_all_ids_from_vulnerability(vuln)
        # if "CVE-2021-23926" in existing_finding_ids or "CVE-2021-23926" in new_finding_ids:
        #     print(f"{existing_finding_ids} new = {new_finding_ids}  in both = {__does_id_exist_in_both(existing_finding_ids, new_finding_ids)}")

        if __does_id_exist_in_both(existing_finding_ids, new_finding_ids):
            return vuln
    return None


def combine_vulnerabilities(
    vulnerabilities_list: List[Dict[str, Optional[List[Vulnerability]]]]
) -> Dict[str, Optional[List[Vulnerability]]]:
    """Intelligently combines the lists of vulnerabilities found by different scanners"""
    combined_result = {}
    for result in vulnerabilities_list:
        for purl in result:
            tool_vulnerabilities = result[purl]
            if purl not in combined_result:
                combined_result[purl] = []

            for vulnerability in tool_vulnerabilities:
                exisiting = __get_matching_vulnerability_from_list(
                    combined_result[purl], vulnerability
                )
                if exisiting is None:
                    combined_result[purl].append(vulnerability)
                else:
                    __combine_vunerabilities(exisiting, vulnerability)
    return combined_result



"""This provides a set of utility functions for working with vulnerabilities and boms"""
# This file is part of hoppr-cop
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Copyright (c) 2022 Lockheed Martin Corporation

import json
import os
import stat
import tempfile
import uuid
from pathlib import Path
from typing import Optional, Union, List, Dict

import requests
from hoppr_cyclonedx_models.cyclonedx_1_3 import (
    CyclonedxSoftwareBillOfMaterialSpecification as Bom_1_3,
)
from hoppr_cyclonedx_models.cyclonedx_1_4 import (
    CyclonedxSoftwareBillOfMaterialsStandard as Bom_1_4,
    VulnerabilitySource,
    Reference,
    Advisory,
)
import typer
from packageurl import PackageURL


def convert_xml_to_json(file_path: Path) -> Path:
    """
    Function to convert a xml file to json format.
    """
    typer.echo("xml format detected, attempt to convert with cyclonedx tools")
    # Default to the path specified in the hoppr-cop docker file or define the local filename to save data
    docker_image_path = Path("/usr/local/bin/cyclone-dx")
    cyclone_dx_path = (
        docker_image_path
        if docker_image_path.exists()
        else Path(tempfile.gettempdir()) / "cyclonedx"
    )
    if not cyclone_dx_path.exists():
        typer.echo("cyclonedx tools not found Attempting to download")
        if "CYCLONE_INSTALL_URL" in os.environ:
            url = os.environ.get("CYCLONE_INSTALL_URL")
            # Make http request for remote file datae
            data = requests.get(url)
            # Save file data to local copy
            with open(cyclone_dx_path, "wb") as file:
                file.write(data.content)
        else:
            msg = typer.style(
                "In order to support xml boms, you must set 'CYCLONE_INSTALL_URL' to "
                "the correct release of cyclone-dx cli. https://github.com/CycloneDX/cyclonedx-cli/releases",
                fg=typer.colors.RED,
            )
            typer.echo(msg)
            raise typer.Exit(code=1)
    os.chmod(
        cyclone_dx_path,
        stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH,
    )
    os.system(
        f"{cyclone_dx_path.absolute()} convert --input-file {file_path} \
            --output-file {tempfile.gettempdir()}/{file_path.name}.json --output-format json"
    )
    return Path(f"{tempfile.gettempdir()}/{file_path.name}.json")


def parse_sbom(sbom_file: Path) -> Optional[Union[Bom_1_4, Bom_1_3]]:
    """Parses a Software Bill of Materials File"""
    sbom = None
    if sbom_file.is_file() and sbom_file.exists():
        typer.echo(f"processing {str(sbom_file)}")
        if str(sbom_file).endswith(".xml"):
            sbom_file = convert_xml_to_json(sbom_file)

        with open(sbom_file, encoding="utf-8") as sbom_file_json:
            sbom_json_object = json.load(sbom_file_json)
            sbom = create_sbom_object(sbom_json_object, str(sbom_file))
    else:
        typer.secho(f"{str(sbom_file)} is not a file", fg=typer.colors.RED)
        raise typer.Exit
    return sbom


def parse_sbom_json_string(
    sbom_file_json: str, sbom_info: str
) -> Optional[Union[Bom_1_4, Bom_1_3]]:
    """Parses a Software Bill of Materials JSON String"""
    sbom_json_object = json.loads(sbom_file_json)
    return create_sbom_object(sbom_json_object, sbom_info)


def create_sbom_object(
    sbom_json_object: Dict, sbom_info: str
) -> Optional[Union[Bom_1_4, Bom_1_3]]:
    """Creates a Software Bill of Materials Object"""
    sbom = None
    spec_version = sbom_json_object.get("specVersion", "")
    if "$schema" in sbom_json_object:
        del sbom_json_object["$schema"]
    if spec_version == "1.4":
        sbom = Bom_1_4(**sbom_json_object)
    elif spec_version == "1.3":
        sbom = Bom_1_3(**sbom_json_object)
    else:
        typer.secho(f"{sbom_info} is an unknown spec version ({spec_version})")
        raise typer.Exit

    return sbom


def get_vulnerability_source(vulnerability_id: str) -> Optional[VulnerabilitySource]:
    """Generate the source for a vulnerability based on a given ID"""
    if vulnerability_id.startswith("CVE-"):
        return VulnerabilitySource(
            name="NVD", url=f"https://nvd.nist.gov/vuln/detail/{vulnerability_id}"
        )
    if vulnerability_id.startswith("GHSA"):
        return VulnerabilitySource(
            name="Github Advisories",
            url=f"https://github.com/advisories/{vulnerability_id}",
        )
    if vulnerability_id.startswith("GMS"):
        return VulnerabilitySource(
            name="Github Advisories",
            url=f"https://github.com/advisories/{vulnerability_id}",
        )
    if vulnerability_id.startswith("sonatype"):
        return VulnerabilitySource(
            name="OSS Index",
            url=f"https://ossindex.sonatype.org/vulnerability/{vulnerability_id}",
        )

    return None
    # else:
    #     efoss_vuln.source = Source(name="Gitlab Advisories",
    #                                url=f"https://gitlab.com/gitlab-org/security-products/gemnasium-db/-/tree/master/{path_slug}/{filename}",
    #                                scanner="gemnasium")


def get_advisories_from_urls(urls: List[str]) -> List[Advisory]:
    """Generates a list of advisories for the given set of urls."""
    urls = list(set(urls))
    return list((map(lambda x: Advisory(url=x), urls)))


def get_references_from_ids(ids: List[str], primary_id: str) -> List[Reference]:
    """Builds a list of Reference objects to the given vulnerability IDs"""
    unique_ids = list(set(ids))
    references = []
    for ident in unique_ids:
        if ident != primary_id:
            references.append(
                Reference(id=ident, source=get_vulnerability_source(ident))
            )
    return references


def build_bom_dict_from_purls(purls: List[PackageURL]) -> Dict:
    bom = {
        "bomFormat": "CycloneDX",
        "specVersion": "1.4",
        "serialNumber": f"urn:uuid:{str(uuid.uuid1())}",
        "version": 1,
        "components": [],
    }

    for purl in purls:
        component = {
            "type": "library",
            "name": purl.name,
            "version": purl.version,
            "purl": purl.to_string(),
            "group": purl.namespace,
            "bom-ref": purl.to_string(),
            "description": "test",
            "author": "test",
            "externalReferences": [],
        }
        if purl.namespace is not None:
            component["group"] = purl.namespace

        bom["components"].append(component)
    return bom


def build_bom_from_purls(purls: List[PackageURL]) -> Bom_1_4:
    """Creates a skeleton bOM from a list of purls"""
    return Bom_1_4(**build_bom_dict_from_purls(purls))



"""Provides vulnerability reporting"""
# This file is part of hoppr-security-commons
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Copyright (c) 2022 Lockheed Martin CorporationØ
import json
import pkgutil
import uuid
from collections import defaultdict
from pathlib import Path
from typing import List, Optional

import jinja2
import typer
from hoppr_cyclonedx_models.cyclonedx_1_4 import (
    CyclonedxSoftwareBillOfMaterialsStandard as Bom_1_4,
)
from hoppr_cyclonedx_models.cyclonedx_1_4 import (
    Vulnerability,
    Affect,
    Rating,
    Advisory,
    ScoreMethod,
    Severity,
)
from packageurl import PackageURL
from tabulate import tabulate

from security_commons.common.reporting.models import ReportFormat


class Reporting:
    """Generates reports in multiple formats from a list of vulnerabilities"""

    output_path: Path
    base_name: str

    def __init__(self, output_path: Path, base_name: str):
        self.output_path = output_path
        self.base_name = base_name

    def generate_vulnerability_reports(
        self,
        formats: List[ReportFormat],
        vulnerabilities: dict[str, Optional[list[Vulnerability]]],
        bom: Optional[Bom_1_4] = None,
    ):
        """Generates various vulnerability reports based on specified formats"""
        flattened_vulnerabilties = self.__add_purl_as_bom_ref_and_flatten(
            vulnerabilities
        )
        if ReportFormat.TABLE in formats:
            findings = self.__get_fields_from_vulnerabilities(flattened_vulnerabilties)
            typer.echo(
                tabulate(
                    findings,
                    headers=["type", "name", "version", "id", "severity", "found by"],
                )
            )
        if ReportFormat.CYCLONE_DX in formats and bom is not None:
            self.add_vulnerabilities_to_bom(bom, flattened_vulnerabilties)
            with open(
                self.output_path / f"{self.base_name}-enhanced.json",
                "w",
                encoding="UTF-8",
            ) as file:
                file.write(bom.json(exclude_none=True, by_alias=True))
                file.close()
        if ReportFormat.HTML in formats:
            self.__generate_html_report(flattened_vulnerabilties)

        if ReportFormat.GITLAB in formats:
            self.generate_gitlab_vulnerability_report(flattened_vulnerabilties)

    @staticmethod
    def add_vulnerabilities_to_bom(
        bom: Bom_1_4, vulnerabilities: List[Vulnerability]
    ) -> Bom_1_4:
        """Adds the vulnerabilities found by various scanners to cyclone 1.4+ compliant BOM"""
        if isinstance(bom, Bom_1_4):
            bom.vulnerabilities = []
            bom.vulnerabilities = bom.vulnerabilities + vulnerabilities
        else:
            typer.echo(
                "Cannot add vulnerabilities to a bom earlier than cyclone version 1.4"
            )
        return bom

    @staticmethod
    def link_vulnerabilities_to_bom(vulnerabilities: List[Vulnerability]) -> Bom_1_4:
        """Creates/Adds the vulnerabilities found by various scanners to cyclone 1.4+ compliant VEX BOM"""
        vex_bom = Bom_1_4(bomFormat="CycloneDX", specVersion="1.4", version=1)
        vex_bom.serialNumber = f"urn:uuid:{uuid.uuid4()}"
        vex_bom.vulnerabilties = vulnerabilities

        return vex_bom

    def __add_purl_as_bom_ref_and_flatten(
        self, vulnerabilities: dict[str, Optional[list[Vulnerability]]]
    ) -> List[Vulnerability]:
        flattened_vulnerabilities: List[Vulnerability] = []

        for purl in vulnerabilities:
            for vuln in vulnerabilities[purl]:
                vuln.affects = [] if vuln.affects is None else vuln.affects
                vuln.affects.append(Affect(**{"ref": purl}))
                flattened_vulnerabilities.append(vuln)

        flattened_vulnerabilities.sort(key=self.get_score, reverse=True)
        return flattened_vulnerabilities

    def get_score(self, vuln_to_score: Vulnerability) -> float:
        best_rating = self.__get_best_rating(vuln_to_score.ratings)
        if best_rating is None or best_rating.score is None:
            return 0.0
        return best_rating.score

    def __get_fields_from_vulnerabilities(self, vulnerabilities: List[Vulnerability]):
        findings = []

        def get_fields(vuln: Vulnerability) -> List:
            tools = [] if vuln.tools is None else vuln.tools
            tools = list(map(lambda x: f"{x.vendor} {x.name}", tools))
            severity = self.__get_severity(vuln.ratings)
            if severity == "critical":
                severity = typer.style(severity, fg=typer.colors.RED)
            elif severity == "high":
                severity = typer.style(severity, fg=typer.colors.BRIGHT_YELLOW)
            purl = PackageURL.from_string(vulnerability.affects[0].ref)
            return [
                purl.type,
                purl.name,
                purl.version,
                vuln.id,
                severity,
                " | ".join(tools),
            ]

        for vulnerability in vulnerabilities:
            findings.append(get_fields(vulnerability))

        return findings

    def __copy_assets(self):
        assets = ["vulnerabilities.css"]
        output_path = self.output_path / "assets"
        output_path.mkdir(exist_ok=True, parents=True)
        for asset in assets:
            output_path = output_path / asset
            template_data = pkgutil.get_data(
                __name__, f"templates/assets/{asset}"
            ).decode("utf-8")
            with open(output_path, "w", encoding="utf-8") as out:
                out.write(template_data)
            out.close()

    def __generate_html_report(self, combined_list: List[Vulnerability]):
        output_path = self.output_path / f"{self.base_name}-vulnerabilities.html"
        template_data = pkgutil.get_data(
            __name__, "templates/vulnerabilities.html"
        ).decode("utf-8")

        env = jinja2.Environment()
        env.filters["severity"] = self.__get_severity
        template = env.from_string(template_data)
        # template = Template(template_data)

        self.__copy_assets()

        severity_classes = {
            "critical": "bg-red-100 rounded-lg py-5 px-6 mb-4 text-base text-red-700 mb-3",
            "high": "bg-yellow-100 rounded-lg py-5 px-6 mb-4 text-base text-yellow-700 mb-3",
            "medium": "bg-gray-50 rounded-lg py-5 px-6 mb-4 text-base text-gray-500 mb-3",
            "info": "bg-gray-50 rounded-lg py-5 px-6 mb-4 text-base text-gray-500 mb-3",
            "low": "bg-gray-50 rounded-lg py-5 px-6 mb-4 text-base text-gray-500 mb-3",
            "unknown": "bg-gray-50 rounded-lg py-5 px-6 mb-4 text-base text-gray-500 mb-3",
            "none": "bg-gray-50 rounded-lg py-5 px-6 mb-4 text-base text-gray-500 mb-3",
        }

        result = template.render(
            {
                "findings": combined_list,
                "severity_classes": severity_classes,
                "base_name": self.base_name,
            }
        )
        with open(output_path, "w", encoding="utf-8") as out:
            out.write(result)
            out.close()
        self.__generate_vuln_detail_reports(combined_list, severity_classes)

    def __generate_vuln_detail_reports(
        self, vulnerabilities: List[Vulnerability], severity_classes
    ):
        output_path = self.output_path / f"{self.base_name}-details"
        output_path.mkdir(exist_ok=True)
        template_data = pkgutil.get_data(
            __name__, "templates/vulnerability_details.html"
        ).decode("utf-8")
        env = jinja2.Environment()
        env.filters["featured_link"] = self.__get_featured_link
        template = env.from_string(template_data)
        for vuln in vulnerabilities:
            purl = PackageURL.from_string(vuln.affects[0].ref)
            result = template.render(
                {
                    "type": purl.type,
                    "namespace": purl.namespace,
                    "name": purl.name,
                    "version": purl.version,
                    "severity_classes": severity_classes,
                    "vulnerability": vuln,
                    "purl": purl.to_string(),
                    "base_name": self.base_name,
                }
            )
            file_name = output_path / f"{vuln.id}.html"
            with open(file_name, "w", encoding="utf-8") as out:
                out.write(result)
                out.close()

    def generate_gitlab_vulnerability_report(
        self, vulnerabilities: List[Vulnerability]
    ):
        """Renders the vulnerabilities report for gitlab"""
        # Note the JSON schema for this report can be found at
        # https://gitlab.com/gitlab-org/security-products/security-report-schemas/-/blob/master/dist/dependency-scanning-report-format.json
        report = {
            "version": "14.1.2",
            "vulnerabilities": [],
            "remediations": [],
            "dependency_files": [],
        }
        dependencies_by_format = defaultdict(lambda: [], {})
        purls = list({vuln.affects[0].ref for vuln in vulnerabilities})
        output_path = self.output_path / "gl-dependency-scanning-report.json"
        for purl in purls:
            purl = PackageURL.from_string(purl)
            dependencies_by_format[purl.type].append(
                {
                    "package": {"name": purl.name},
                    "version": purl.version,
                }
            )

        for vuln in vulnerabilities:
            report["vulnerabilities"].append(self.__generate_gitlab_row(vuln))

        for repo_format in dependencies_by_format:
            report["dependency_files"].append(
                {
                    "package_manager": repo_format,
                    "path": "cyclonedx.bom",
                    "dependencies": dependencies_by_format[repo_format],
                }
            )

        with open(output_path, "w", encoding="utf-8") as out:
            out.write(json.dumps(report, indent=4, sort_keys=True, default=str))
            out.close()

    @staticmethod
    def __get_featured_link(advisories: Optional[List[Advisory]]) -> Optional[str]:
        if advisories is not None:
            for adv in advisories:
                url = "" if adv.url is None else adv.url
                if "https://snyk.io/" in url:
                    return url
        return None

    def __get_severity(self, ratings: Optional[List[Rating]]) -> str:
        best_rating = self.__get_best_rating(ratings)
        if best_rating is None or best_rating.severity is None:
            return "none"

        return Severity(best_rating.severity).value

    @staticmethod
    def __get_best_rating(ratings: Optional[List[Rating]]):
        default_rating = None if ratings is None or len(ratings) == 0 else ratings[0]
        methods = [
            ScoreMethod(rating.method).value if rating.method is not None else "none"
            for rating in ratings
        ]
        preferred_method = None
        if "CVSSv31" in methods:
            preferred_method = "CVSSv31"
        elif "CVSSv3" in methods:
            preferred_method = "CVSSv3"
        elif "CVSSv2" in methods:
            preferred_method = "CVSSv2"

        return next(
            filter(
                lambda rating: rating.method
                and ScoreMethod(rating.method).value == preferred_method,
                ratings,
            ),
            default_rating,
        )

    def __generate_gitlab_row(self, vuln: Vulnerability):
        """Generates a report row"""
        purl = PackageURL.from_string(vuln.affects[0].ref)
        severity = self.__get_severity(vuln.ratings).title()
        return {
            "category": "dependency_scanning",
            "name": vuln.description,
            "description": vuln.description,
            "cve": vuln.id,
            "severity": severity if severity != "none" else "Info",
            "confidence": "Unknown",
            "identifiers": [{"type": "cve", "name": vuln.id, "value": vuln.id}],
            "scanner": {"id": vuln.tools[0].name, "name": vuln.tools[0].name},
            "location": {
                "file": "cyclonedx.bom",
                "dependency": {"package": {"name": purl.name}, "version": purl.version},
            },
        }
